#  Copyright 2000-2023 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
import re
import sys
import os
import time
from contextlib import contextmanager

if sys.version_info[0] == 2:
    from StringIO import StringIO
else:
    from io import StringIO

from pydev_tests_python.debugger_unittest import DebuggerRunner, AbstractWriterThread, \
    start_in_daemon_thread, read_process, SHOW_OTHER_DEBUG_INFO, wait_for_condition

TIMEOUT = 2
TRIM_PATTERN = '('
TEMP_FILES_DIR = 'resources'
RESULT_FILE = 'error_lines.txt'
exception_types = ('ImportError', 'AttributeError', 'ModuleNotFoundError', 'NameError')
HELP_MODULES_START_STRING = 'Please wait a moment while I gather a list of all available modules...'
HELP_MODULES_END_STRING = 'Enter any module name to get more help.  Or, type "modules spam" to search for modules whose name or summary contain the string "spam".'
HELP_MODULES_END_STRING_PY2 = 'Enter any module name to get more help.  Or, type "modules spam" to search for modules whose descriptions contain the word "spam".'

temp_file_dir_path = os.path.join(os.path.dirname(__file__), TEMP_FILES_DIR)
resulting_dict = dict()


def get_builtin_module_path(file_name):
    try:
        r_path = os.path.realpath  # @UndefinedVariable
    except:
        r_path = os.path.abspath

    ret = os.path.normcase(r_path(os.path.join(os.path.dirname(__file__), file_name)))
    if not os.path.exists(ret):
        ret = os.path.join(os.path.dirname(ret), TEMP_FILES_DIR, os.path.basename(ret))
    if not os.path.exists(ret):
        raise AssertionError('Expected: %s to exist.' % (ret,))
    return ret


def create_temp_dir():
    try:
        os.mkdir(temp_file_dir_path)
    except:
        pass


def create_temp_file(file_name):
    open(os.path.join(temp_file_dir_path, file_name), 'w')


def clear_temp_dir():
    try:
        for file_name in os.listdir(temp_file_dir_path):
            if file_name.startswith("__py"):
                continue
            os.remove(os.path.join(temp_file_dir_path, file_name))
    except Exception as e:
        print(e)


def get_builtin_modules_name():
    old_stdout = sys.stdout
    result = StringIO()
    sys.stdout = result
    help('modules')
    sys.stdout = old_stdout
    result = result.getvalue().replace('\n\n', ' ').replace('\n', ' ')
    result = result.replace(HELP_MODULES_START_STRING, '')
    result = result.replace(HELP_MODULES_END_STRING, '')
    result = result.replace(HELP_MODULES_END_STRING_PY2, '')
    result = re.split("\s+", result)
    return list(filter(lambda s: len(s) > 0 and not s.startswith('pydev') and not s.startswith('_pydev'), result))


def update_environ():
    os.environ['PWD'] = temp_file_dir_path
    os.environ['PYTHONPATH'] = os.environ['PYTHONPATH'] + os.pathsep + temp_file_dir_path
    os.environ['IDE_PROJECT_ROOTS'] = temp_file_dir_path
    os.chdir(temp_file_dir_path)


def write_result_to_file():
    with open(os.path.join(os.path.dirname(__file__), RESULT_FILE), 'a') as file:
        for key, val in resulting_dict.items():
            file.write('"%s" to "%s",\n' % (key, val,))
    resulting_dict.clear()


def parse_output(file_name, err_massage):
    if err_massage is None:
        return
    lines = err_massage.splitlines()
    for line in lines:
        for exception_type in exception_types:
            if line.startswith(exception_type):
                ind = line.find(TRIM_PATTERN)
                if ind != -1:
                    line = line[:ind]
                line = line.strip()
                resulting_dict[line] = file_name


def debug_file(file_name):
    try:
        with generate_module_error_string().gen(file_name) as w:
            w.write_make_initial_run()
            w.finished_ok = True
            writer = w
        while not writer.stop_working:
            time.sleep(.2)
    except AssertionError:
        pass
    finally:
        if writer is not None:
            parse_output(file_name, writer.error_string)


class GenerationDebuggerRunner(DebuggerRunner):
    def get_command_line(self):
        return [sys.executable, '-u']

    def fail_with_message(self, msg, stdout, stderr, writerThread):
        raise AssertionError(msg)

    @contextmanager
    def check_case(self, writer_class):
        if callable(writer_class):
            writer = writer_class()
        else:
            writer = writer_class

        update_environ()

        try:
            writer.start()
            wait_for_condition(lambda: hasattr(writer, 'port'), timeout=TIMEOUT)

            self.writer = writer
            args = self.get_command_line()
            args = self.add_command_line_args(args)

            with self.run_process(args, writer) as dct_with_stdout_stderr:
                try:
                    wait_for_condition(lambda: writer.finished_initialization,
                                       timeout=TIMEOUT)
                except:
                    writer.finished_initialization = True
                finally:
                    writer.get_stdout = lambda: ''.join(dct_with_stdout_stderr['stdout'])
                    writer.get_stderr = lambda: ''.join(dct_with_stdout_stderr['stderr'])

                yield writer
        except:
            pass
        finally:
            writer.do_kill()
            writer.log = []

        stdout = dct_with_stdout_stderr['stdout']
        stderr = dct_with_stdout_stderr['stderr']
        writer.additional_output_checks(''.join(stdout), ''.join(stderr))

    @contextmanager
    def run_process(self, args, writer):
        process = self.create_process(args, writer)
        stdout = []
        stderr = []
        finish = [False]
        dct_with_stdout_stder = {}

        try:
            start_in_daemon_thread(read_process, (
                process.stdout, stdout, sys.stdout, 'stdout', finish))
            start_in_daemon_thread(read_process, (
                process.stderr, stderr, sys.stderr, 'stderr', finish))

            if SHOW_OTHER_DEBUG_INFO:
                print('Both processes started')

            initial_time = time.time()

            dct_with_stdout_stder['stdout'] = stdout
            dct_with_stdout_stder['stderr'] = stderr
            yield dct_with_stdout_stder

            if not writer.finished_ok:
                self.fail_with_message(
                    "The thread that was doing the tests didn't finish successfully.",
                    stdout, stderr, writer)

            while True:
                if process.poll() is not None:
                    break
                else:
                    if writer is not None:
                        if writer.FORCE_KILL_PROCESS_WHEN_FINISHED_OK:
                            process.kill()
                            continue
                        if time.time() - initial_time > TIMEOUT:  # timed out
                            process.kill()
                            break
                time.sleep(.2)
        except TimeoutError:
            writer.write_dump_threads()
            time.sleep(.2)
            raise
        finally:
            finish[0] = True


def generate_module_error_string():
    runner = GenerationDebuggerRunner()

    class WriterThread(AbstractWriterThread):
        def __init__(self, *args, **kwargs):
            AbstractWriterThread.__init__(self, *args, **kwargs)
            self.error_string = None
            self.stop_working = False

        def additional_output_checks(self, stdout, stderr):
            self.error_string = stderr
            self.stop_working = True

    class Generator(object):
        @contextmanager
        def gen(self, filename, **kwargs):
            WriterThread.TEST_FILE = get_builtin_module_path(filename)
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)
            try:
                with runner.check_case(WriterThread) as writer:
                    yield writer
            except:
                pass

    return Generator()



# -*- coding: utf-8 -*-
"""
Qt6's inputhook support function

Author: Christian Boos
"""

#-----------------------------------------------------------------------------
#  Copyright (C) 2011  The IPython Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import os
import signal

import threading


from PyQt6 import QtCore, QtGui
from pydev_ipython.inputhook import allow_CTRL_C, ignore_CTRL_C, stdin_ready

# To minimise future merging complexity, rather than edit the entire code base below
# we fake InteractiveShell here
class InteractiveShell:
    _instance = None
    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance
    def set_hook(self, *args, **kwargs):
        # We don't consider the pre_prompt_hook because we don't have
        # KeyboardInterrupts to consider since we are running under PyDev
        pass


#-----------------------------------------------------------------------------
# Module Globals
#-----------------------------------------------------------------------------

got_kbdint = False
sigint_timer = None

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------

def create_inputhook_qt6(mgr, app=None):
    """Create an input hook for running the Qt6 application event loop.

    Parameters
    ----------
    mgr : an InputHookManager

    app : Qt Application, optional.
        Running application to use.  If not given, we probe Qt for an
        existing application object, and create a new one if none is found.

    Returns
    -------
    A pair consisting of a Qt Application (either the one given or the
    one found or created) and a inputhook.

    Notes
    -----
    We use a custom input hook instead of PyQt6's default one, as it
    interacts better with the readline packages (issue #481).

    The inputhook function works in tandem with a 'pre_prompt_hook'
    which automatically restores the hook as an inputhook in case the
    latter has been temporarily disabled after having intercepted a
    KeyboardInterrupt.
    """

    if app is None:
        app = QtCore.QCoreApplication.instance()
        if app is None:
            from PyQt6 import QtWidgets
            app = QtWidgets.QApplication([" "])

    # Re-use previously created inputhook if any
    ip = InteractiveShell.instance()
    if hasattr(ip, '_inputhook_qt6'):
        return app, ip._inputhook_qt6

    # Otherwise create the inputhook_qt6/preprompthook_qt6 pair of
    # hooks (they both share the got_kbdint flag)

    def inputhook_qt6():
        """PyOS_InputHook python hook for Qt6.

        Process pending Qt events and if there's no pending keyboard
        input, spend a short slice of time (50ms) running the Qt event
        loop.

        As a Python ctypes callback can't raise an exception, we catch
        the KeyboardInterrupt and temporarily deactivate the hook,
        which will let a *second* CTRL+C be processed normally and go
        back to a clean prompt line.
        """
        try:
            allow_CTRL_C()
            app = QtCore.QCoreApplication.instance()
            if not app: # shouldn't happen, but safer if it happens anyway...
                return 0
            app.processEvents(QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 300)
            if not stdin_ready():
                # Generally a program would run QCoreApplication::exec()
                # from main() to enter and process the Qt event loop until
                # quit() or exit() is called and the program terminates.
                #
                # For our input hook integration, we need to repeatedly
                # enter and process the Qt event loop for only a short
                # amount of time (say 50ms) to ensure that Python stays
                # responsive to other user inputs.
                #
                # A naive approach would be to repeatedly call
                # QCoreApplication::exec(), using a timer to quit after a
                # short amount of time. Unfortunately, QCoreApplication
                # emits an aboutToQuit signal before stopping, which has
                # the undesirable effect of closing all modal windows.
                #
                # To work around this problem, we instead create a
                # QEventLoop and call QEventLoop::exec(). Other than
                # setting some state variables which do not seem to be
                # used anywhere, the only thing QCoreApplication adds is
                # the aboutToQuit signal which is precisely what we are
                # trying to avoid.
                timer = QtCore.QTimer()
                event_loop = QtCore.QEventLoop()
                timer.timeout.connect(event_loop.quit)
                while not stdin_ready():
                    timer.start(50)
                    # Warning: calling event_loop.exec_() can lead to hangs in REPL on mscOS PY-31931
                    # Replacing it with event_loop.processEvents() fixes the issue, but leads to high CPU load on every os PY-42688
                    event_loop.exec()
                    timer.stop()
        except KeyboardInterrupt:
            global got_kbdint, sigint_timer

            ignore_CTRL_C()
            got_kbdint = True
            mgr.clear_inputhook()

            # This generates a second SIGINT so the user doesn't have to
            # press CTRL+C twice to get a clean prompt.
            #
            # Since we can't catch the resulting KeyboardInterrupt here
            # (because this is a ctypes callback), we use a timer to
            # generate the SIGINT after we leave this callback.
            #
            # Unfortunately this doesn't work on Windows (SIGINT kills
            # Python and CTRL_C_EVENT doesn't work).
            if(os.name == 'posix'):
                pid = os.getpid()
                if(not sigint_timer):
                    sigint_timer = threading.Timer(.01, os.kill,
                                                   args=[pid, signal.SIGINT] )
                    sigint_timer.start()
            else:
                print("\nKeyboardInterrupt - Ctrl-C again for new prompt")


        except: # NO exceptions are allowed to escape from a ctypes callback
            ignore_CTRL_C()
            from traceback import print_exc
            print_exc()
            print("Got exception from inputhook_qt6, unregistering.")
            mgr.clear_inputhook()
        finally:
            allow_CTRL_C()
        return 0

    def preprompthook_qt6(ishell):
        """'pre_prompt_hook' used to restore the Qt6 input hook

        (in case the latter was temporarily deactivated after a
        CTRL+C)
        """
        global got_kbdint, sigint_timer

        if(sigint_timer):
            sigint_timer.cancel()
            sigint_timer = None

        if got_kbdint:
            mgr.set_inputhook(inputhook_qt6)
        got_kbdint = False

    ip._inputhook_qt6 = inputhook_qt6
    ip.set_hook('pre_prompt_hook', preprompthook_qt6)

    return app, inputhook_qt6



#  Copyright 2000-2021 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.

import sys
from IPython.core.displayhook import DisplayHook
from IPython.core.displaypub import DisplayPublisher


class PyDevDisplayHook(DisplayHook):
    """ Used for execution result """
    def write_format_data(self, format_dict, *args, **kwargs):
        if not is_supported(format_dict):
            super(PyDevDisplayHook, self).write_format_data(format_dict, *args,
                                                            **kwargs)
            return
        add_new_line_to_text(format_dict)
        format_dict["execution_count"] = str(get_current_exec_count(self.shell))
        send_rich_output(format_dict)


class PyDevDisplayPub(DisplayPublisher):
    """ Used for display() function """
    def publish(self, data, *args, **kwargs):
        if not is_supported(data):
            super(PyDevDisplayPub, self).publish(data, *args, **kwargs)
            return
        add_new_line_to_text(data)
        data["execution_count"] = str(get_current_exec_count(self.shell))
        send_rich_output(data)


def get_current_exec_count(shell):
    if hasattr(shell, "pydev_curr_exec_line") and shell.pydev_curr_exec_line != 0:
        return shell.pydev_curr_exec_line
    else:
        return shell.execution_count


def add_new_line_to_text(format_dict):
    if "text/plain" in format_dict:
        text = format_dict["text/plain"]
        if len(text) > 0 and not text.endswith("\n"):
            format_dict["text/plain"] = text + "\n"


def is_supported(data):
    for type in data.keys():
        if type not in ("text/plain", "image/png", "text/html"):
            return False
        if type == "text/html":
            html = data["text/html"]
            if not is_data_frame(html):
                return False
    return True


def is_data_frame(html):
    return ("javascript" not in html) and ("dataframe" in html) and ("<table" in html)


def send_rich_output(data):
    if 'image/png' in data:
        png = data['image/png']
        if not isinstance(png, str):
            import base64
            res = base64.b64encode(png)
            data['image/png'] = res
    from _pydev_bundle.pydev_ipython_console_011 import get_client
    client = get_client()
    if client:
        client.sendRichOutput(data)
    else:
        sys.stderr.write("Client is None!\n")
        sys.stderr.flush()


def patch_stdout(pydev_shell):
    sys.stdout = PydevStdOut(sys.stdout, pydev_shell)


class PydevStdOut:
    def __init__(self, original_stdout=sys.stdout, pydev_shell=None, *args, **kwargs):
        self.encoding = sys.stdout.encoding
        self.original_stdout = original_stdout
        self.pydev_shell = pydev_shell

    def write(self, s):
        data = {'text/plain': s}
        data['execution_count'] = str(self.pydev_shell.execution_count)
        send_rich_output(data)

    def __getattr__(self, item):
        # it's called if the attribute wasn't found
        if hasattr(self.original_stdout, item):
            return getattr(self.original_stdout, item)
        raise AttributeError("%s has no attribute %s" % (self.original_stdout, item))




#  Copyright 2000-2023 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
import io
import numpy as np

TABLE_TYPE_NEXT_VALUE_SEPARATOR = '__pydev_table_column_type_val__'
MAX_COLWIDTH = 100000

ONE_DIM, TWO_DIM, WITH_TYPES = range(3)
NP_ROWS_TYPE = "int64"

CSV_FORMAT_SEPARATOR = '~'

is_pd = False
try:
    import pandas as pd
    is_pd = True
except:
    pass


def get_type(table):
    # type: (np.ndarray) -> str
    return str(type(table))


def get_shape(table):
    # type: (np.ndarray) -> str
    if table.dtype.names is not None:
        return str((table.shape[0], len(table.dtype.names)))
    if table.ndim == 1:
        return str((table.shape[0], 1))
    elif table.ndim == 0:
        return str((0, 0))
    else:
        return str((table.shape[0], table.shape[1]))


def get_head(table):
    # type: (np.ndarray) -> str
    column_names = table.dtype.names
    if column_names:
        return TABLE_TYPE_NEXT_VALUE_SEPARATOR.join([str(column_names[i]) for i in range(len(column_names))])
    return "None"


def get_column_types(table):
    # type: (np.ndarray) -> str
    if table.ndim == 0:
        return ""
    table = __create_table(table[:1])
    try:
        cols_types = [str(t) for t in table.dtypes] if is_pd else table.get_cols_types()
    except AttributeError:
        cols_types = table.get_cols_types()

    return NP_ROWS_TYPE + TABLE_TYPE_NEXT_VALUE_SEPARATOR + \
        TABLE_TYPE_NEXT_VALUE_SEPARATOR.join(cols_types)


def get_data(table, use_csv_serialization, start_index=None, end_index=None, format=None):
    # type: (Union[np.ndarray, dict], int, int) -> str
    def convert_data_to_html(data, format):
        return repr(__create_table(data, start_index, end_index, format).to_html(notebook=True))

    def convert_data_to_csv(data, format):
        return repr(__create_table(data, start_index, end_index, format).to_csv(na_rep ="None", float_format=format, sep=CSV_FORMAT_SEPARATOR))

    if use_csv_serialization:
        computed_data = __compute_data(table, convert_data_to_csv, format)
    else:
        computed_data = __compute_data(table, convert_data_to_html, format)
    return computed_data


def display_data_html(table, start_index=None, end_index=None):
    # type: (np.ndarray, int, int) -> None
    def ipython_display(data, format):
        from IPython.display import display, HTML
        display(HTML(__create_table(data, start_index, end_index).to_html(notebook=True)))

    __compute_data(table, ipython_display)


def display_data_csv(table, start_index=None, end_index=None):
    # type: (np.ndarray, int, int) -> None
    def ipython_display(data, format):
        print(repr(__create_table(data, start_index, end_index).to_csv(na_rep ="None", sep=CSV_FORMAT_SEPARATOR, float_format=format)))

    __compute_data(table, ipython_display)


def remove_nones_values(array_part, na_rep):
    if np.issubdtype(array_part.dtype, np.number):
        array_part_without_nones = np.where(array_part == None, np.nan, array_part)
    else:
        array_part_without_nones = np.where(array_part == None, na_rep, array_part)
    return array_part_without_nones


class _NpTable:
    def __init__(self, np_array, format=None):
        self.array = np_array
        self.type = self.get_array_type()
        self.indexes = None
        self.format = format

    def get_array_type(self):
        col_type = self.array.dtype

        if len(col_type) != 0:
            return WITH_TYPES

        if self.array.ndim > 1:
            return TWO_DIM

        return ONE_DIM

    def get_cols_types(self):
        col_type = self.array.dtype

        if self.type == ONE_DIM:
            # [1, 2, 3] -> [int]
            return [str(col_type)]

        if self.type == WITH_TYPES:
            # ([(10, 3.14), (20, 2.71)], dtype=[("ci", "i4"), ("cf", "f4")]) -> [int, float]
            return [str(col_type[i]) for i in range(len(col_type))]  # is not iterable

        # [[1, 2], [3, 4]] -> [int, int]
        return [str(col_type) for _ in range(len(self.array[0]))]

    def head(self, num_rows):
        if self.array.shape[0] < 6:
            return self

        return _NpTable(self.array[:5]).sort()

    def to_html(self, notebook):
        html = ['<table class="dataframe">\n']

        # columns names
        html.append('<thead>\n'
                    '<tr style="text-align: right;">\n'
                    '<th></th>\n')
        html += self.__collect_cols_names()
        html.append('</tr>\n'
                    '</thead>\n')

        # tbody
        html += self.__collect_values(None)

        html.append('</table>\n')

        return "".join(html)

    def __collect_cols_names(self):
        if self.type == ONE_DIM:
            return ['<th>0</th>\n']

        if self.type == WITH_TYPES:
            columns_names = self.array.dtype.names
            return ['<th>{}</th>\n'.format(str(columns_names[i])) for i in range(len(columns_names))]

        return ['<th>{}</th>\n'.format(i) for i in range(len(self.array[0]))]

    def __collect_values(self, max_cols):
        html = ['<tbody>\n']
        rows = self.array.shape[0]
        for row_num in range(rows):
            html.append('<tr>\n')
            html.append('<th>{}</th>\n'.format(int(self.indexes[row_num])))
            if self.type == ONE_DIM:
                if self.format is not None and self.array[row_num] is not None and self.array[row_num] == self.array[row_num]:
                    try:
                        value = self.format % self.array[row_num]
                    except Exception as _:
                        value = self.array[row_num]
                else:
                    value = self.array[row_num]
                html.append('<td>{}</td>\n'.format(value))
            else:
                cols = len(self.array[0])
                max_cols = cols if max_cols is None else min(max_cols, cols)
                for col_num in range(max_cols):
                    if self.format is not None and self.array[row_num][col_num] is not None and self.array[row_num][col_num] == self.array[row_num][col_num]:
                        try:
                            value = self.format % self.array[row_num][col_num]
                        except Exception as _:
                            value = self.array[row_num][col_num]
                    else:
                        value = self.array[row_num][col_num]
                    html.append('<td>{}</td>\n'.format(value))
            html.append('</tr>\n')
        html.append('</tbody>\n')
        return html

    def to_csv(self, na_rep="None", float_format=None, sep=CSV_FORMAT_SEPARATOR):
        csv_stream = io.StringIO()
        if self.array.dtype.names is not None:
            np_array_without_nones = []
            for field in self.array.dtype.names:
                np_array_without_nones.append(remove_nones_values(self.array[str(field)], na_rep))
            np_array_without_nones = np.column_stack(np_array_without_nones)
        else:
            np_array_without_nones = remove_nones_values(self.array, na_rep)
        if float_format is None or float_format == 'null':
            float_format = "%s"

        np.savetxt(csv_stream, np_array_without_nones, delimiter=sep, fmt=float_format)
        csv_string = csv_stream.getvalue()
        csv_rows_with_index = self.__insert_index_at_rows_begging_csv(csv_string)

        col_names = self.__collect_col_names_csv()
        return col_names + "\n" + csv_rows_with_index

    def __insert_index_at_rows_begging_csv(self, csv_string):
        # type: (str) -> str
        csv_rows = csv_string.split('\n')
        csv_rows_with_index = []
        for row_index in range(self.array.shape[0]):
            csv_rows_with_index.append(str(row_index) + CSV_FORMAT_SEPARATOR + csv_rows[row_index])
        return "\n".join(csv_rows_with_index)

    def __collect_col_names_csv(self):
        if self.type == ONE_DIM:
            return '{}0'.format(CSV_FORMAT_SEPARATOR)

        if self.type == WITH_TYPES:
            return CSV_FORMAT_SEPARATOR + CSV_FORMAT_SEPARATOR.join(['{}'.format(name) for name in self.array.dtype.names])

        # TWO_DIM
        return CSV_FORMAT_SEPARATOR + CSV_FORMAT_SEPARATOR.join(['{}'.format(i) for i in range(self.array.shape[1])])


    def slice(self, start_index=None, end_index=None):
        if end_index is not None and start_index is not None:
            self.array = self.array[start_index:end_index]
            self.indexes = self.indexes[start_index:end_index]

        return self

    def sort(self, sort_keys=None):
        self.indexes = np.arange(self.array.shape[0])
        if sort_keys is None:
            return self

        cols, orders = sort_keys
        if 0 in cols:
            return self.__sort_by_index(True in orders)

        if self.type == ONE_DIM:
            extended = np.column_stack((self.indexes, self.array))
            sort_extended = extended[:, 1].argsort()
            if False in orders:
                sort_extended = sort_extended[::-1]
            result = extended[sort_extended]
            self.array = result[:, 1]
            self.indexes = result[:, 0]
            return self

        if self.type == WITH_TYPES:
            new_dt = np.dtype([('_pydevd_i', 'i8')] + self.array.dtype.descr)
            extended = np.zeros(self.array.shape, dtype=new_dt)
            extended['_pydevd_i'] = list(range(self.array.shape[0]))
            for col in self.array.dtype.names:
                extended[col] = self.array[col]

            column_names = self.array.dtype.names
            for i in range(len(cols) - 1, -1, -1):
                name = column_names[cols[i] - 1]
                sort = extended[name].argsort(kind='stable')
                extended = extended[sort if orders[i] else sort[::-1]]
            self.indexes = extended['_pydevd_i']
            for col in self.array.dtype.names:
                self.array[col] = extended[col]
            return self

        extended = np.insert(self.array, 0, self.indexes, axis=1)
        for i in range(len(cols) - 1, -1, -1):
            sort = extended[:, cols[i]].argsort(kind='stable')
            extended = extended[sort if orders[i] else sort[::-1]]
        self.indexes = extended[:, 0]
        self.array = extended[:, 1:]
        return self

    def __sort_by_index(self, order):
        if order:
            return self
        self.array = self.array[::-1]
        self.indexes = self.indexes[::-1]
        return self


def __sort_df(dataframe, sort_keys):
    if sort_keys is None:
        return dataframe

    cols, orders = sort_keys
    if 0 in cols:
        if len(cols) == 1:
            return dataframe.sort_index(ascending=orders[0])
        return dataframe.sort_index(level=cols, ascending=orders)
    sort_by = list(map(lambda c: dataframe.columns[c - 1], cols))
    return dataframe.sort_values(by=sort_by, ascending=orders)


def __create_table(command, start_index=None, end_index=None, format=None):
    sort_keys = None

    if type(command) is dict:
        np_array = command['data']
        sort_keys = command['sort_keys']
    else:
        np_array = command

    if is_pd:
        sorted_df = __sort_df(pd.DataFrame(np_array), sort_keys)
        if start_index is not None and end_index is not None:
            sorted_df_slice = sorted_df.iloc[start_index:end_index]
            # to apply "format" we should not have None inside DFs
            try:
                import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    sorted_df_slice = sorted_df_slice.fillna("None")
            except Exception as _:
                pass
            return sorted_df_slice
        return sorted_df

    return _NpTable(np_array, format=format).sort(sort_keys).slice(start_index,
                                                                   end_index)


def __compute_data(arr, fun, format=None):
    is_sort_command = type(arr) is dict
    data = arr['data'] if is_sort_command else arr

    jb_max_cols, jb_max_colwidth, jb_max_rows, jb_float_options = None, None, None, None
    if is_pd:
        jb_max_cols, jb_max_colwidth, jb_max_rows, jb_float_options = __set_pd_options(format)

    if is_sort_command:
        arr['data'] = data
        data = arr

    format = pd.get_option('display.float_format') if is_pd else format

    data = fun(data, format)

    if is_pd:
        __reset_pd_options(jb_max_cols, jb_max_colwidth, jb_max_rows, jb_float_options)

    return data


def __get_tables_display_options():
    # type: () -> Tuple[None, Union[int, None], None]
    try:
        import pandas as pd
        # In pandas versions earlier than 1.0, max_colwidth must be set as an integer
        if int(pd.__version__.split('.')[0]) < 1:
            return None, MAX_COLWIDTH, None
    except Exception:
        pass
    return None, None, None


def __set_pd_options(format):
    max_cols, max_colwidth, max_rows = __get_tables_display_options()
    _jb_float_options = None

    _jb_max_cols = pd.get_option('display.max_columns')
    _jb_max_colwidth = pd.get_option('display.max_colwidth')
    _jb_max_rows = pd.get_option('display.max_rows')
    if format is not None:
        _jb_float_options = pd.get_option('display.float_format')

    pd.set_option('display.max_columns', max_cols)
    pd.set_option('display.max_rows', max_rows)
    pd.set_option('display.max_colwidth', max_colwidth)
    format_function = __define_format_function(format)
    if format_function is not None:
        pd.set_option('display.float_format', format_function)

    return _jb_max_cols, _jb_max_colwidth, _jb_max_rows, _jb_float_options


def __reset_pd_options(max_cols, max_colwidth, max_rows, float_format):
    pd.set_option('display.max_columns', max_cols)
    pd.set_option('display.max_colwidth', max_colwidth)
    pd.set_option('display.max_rows', max_rows)
    if float_format is not None:
        pd.set_option('display.float_format', float_format)


def __define_format_function(format):
    # type: (Union[None, str]) -> Union[Callable, None]
    if format is None or format == 'null':
        return None

    if type(format) == str and format.startswith("%"):
        return lambda x: format % x
    else:
        return None



from _pydevd_bundle import pydevd_thrift
from _pydevd_bundle.pydevd_comm import PyDBDaemonThread
from _pydevd_bundle.pydevd_comm import threading
from _pydevd_bundle.pydevd_comm import time
from _pydevd_bundle.pydevd_constants import ASYNC_EVAL_TIMEOUT_SEC


class ThriftAbstractGetValueAsyncThread(PyDBDaemonThread):
    """
    Abstract class for a thread, which evaluates values for async variables
    """
    def __init__(self, server, seq, var_objects, user_type_renderers=None):
        PyDBDaemonThread.__init__(self)
        self.server = server
        self.seq = seq
        self.var_objs = var_objects
        self.cancel_event = threading.Event()
        self.user_type_renderers = user_type_renderers

    def send_result(self, xml):
        raise NotImplementedError()

    def _on_run(self):
        start = time.time()
        values = []
        for (var_obj, name) in self.var_objs:
            current_time = time.time()
            if current_time - start > ASYNC_EVAL_TIMEOUT_SEC or self.cancel_event.is_set():
                break
            # pydev_console_thrift.DebugValue()
            values.append(pydevd_thrift.var_to_struct(var_obj, name, evaluate_full_value=True, user_type_renderers=self.user_type_renderers))
        self.send_result(values)


class ThriftGetValueAsyncThreadConsole(ThriftAbstractGetValueAsyncThread):
    """
    A thread for evaluation async values, which returns result for Console
    Send result directly to Console's server
    """
    def send_result(self, values):
        if self.server is not None:
            self.server.returnFullValue(self.seq, values)



#  Copyright 2000-2025 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
from _pydevd_bundle.tables.images.pydevd_image_loader import save_image_to_storage, DEFAULT_IMAGE_FORMAT


def create_image(figure):
    # type: (Union[matplotlib.figure.Figure | plotly.graph_objs._figure.Figure]) -> str
    try:
        try:
            import matplotlib.figure
        except ImportError:
            matplotlib = None

        try:
            from plotly.graph_objects import Figure as PlotlyFigure
        except ImportError:
            PlotlyFigure = None

        if matplotlib and isinstance(figure, matplotlib.figure.Figure):
            return save_image_to_storage(figure, format=DEFAULT_IMAGE_FORMAT, save_func=lambda buffer, fmt: figure.savefig(buffer, format=fmt))
        elif PlotlyFigure and isinstance(figure, PlotlyFigure):
            return save_image_to_storage(figure, format=DEFAULT_IMAGE_FORMAT, save_func=lambda buffer, fmt: buffer.write(figure.to_image(format=fmt)))
        else:
            return "Error: Unsupported figure type."
    except Exception as e:
        return "Error: {}".format(e)



class CodeFragment:
    def __init__(self, text, is_single_line=True):
        self.text = text
        self.is_single_line = is_single_line

    def append(self, code_fragment):
        self.text = self.text + "\n" + code_fragment.text
        if not code_fragment.is_single_line:
            self.is_single_line = False


class Command:
    def __init__(self, interpreter, code_fragment):
        """
        :type code_fragment: CodeFragment
        :type interpreter: InteractiveConsole
        """
        self.interpreter = interpreter
        self.code_fragment = code_fragment
        self.more = None
        self.exception_occurred = False

    @staticmethod
    def symbol_for_fragment(code_fragment):
        if code_fragment.is_single_line:
            symbol = 'single'
        else:
            symbol = 'exec' # Jython doesn't support this
        return symbol

    def run(self):
        text = self.code_fragment.text
        symbol = self.symbol_for_fragment(self.code_fragment)

        self.more = self.interpreter.runsource(text, '<input>', symbol)



def get_main_thread_instance(threading):
    if hasattr(threading, "main_thread"):
        return threading.main_thread()
    else:
        # On Python 2 we don't really have an API to get the main thread,
        # so, we just get it from the 'shutdown' bound method.
        return threading._shutdown.im_self


def get_main_thread_id(unlikely_thread_id=None):
    """
    :param unlikely_thread_id:
        Pass to mark some thread id as not likely the main thread.

    :return tuple(thread_id, critical_warning)
    """
    import sys
    import os

    current_frames = sys._current_frames()
    possible_thread_ids = []
    for thread_ident, frame in current_frames.items():
        while frame.f_back is not None:
            frame = frame.f_back

        basename = os.path.basename(frame.f_code.co_filename)
        if basename.endswith((".pyc", ".pyo")):
            basename = basename[:-1]

        if (frame.f_code.co_name, basename) in [
            ("_run_module_as_main", "runpy.py"),
            ("_run_module_as_main", "<frozen runpy>"),
            ("run_module_as_main", "runpy.py"),
            ("run_module", "runpy.py"),
            ("run_path", "runpy.py"),
        ]:
            # This is the case for python -m <module name> (this is an ideal match, so,
            # let's return it).
            return thread_ident, ""

        if frame.f_code.co_name == "<module>":
            if frame.f_globals.get("__name__") == "__main__":
                possible_thread_ids.insert(0, thread_ident)  # Add with higher priority
                continue

            # Usually the main thread will be started in the <module>, whereas others would
            # be started in another place (but when Python is embedded, this may not be
            # correct, so, just add to the available possibilities as we'll have to choose
            # one if there are multiple).
            possible_thread_ids.append(thread_ident)

    if len(possible_thread_ids) > 0:
        if len(possible_thread_ids) == 1:
            return possible_thread_ids[0], ""  # Ideal: only one match

        while unlikely_thread_id in possible_thread_ids:
            possible_thread_ids.remove(unlikely_thread_id)

        if len(possible_thread_ids) == 1:
            return possible_thread_ids[0], ""  # Ideal: only one match

        elif len(possible_thread_ids) > 1:
            # Bad: we can't really be certain of anything at this point.
            return possible_thread_ids[0], "Multiple thread ids found (%s). Choosing main thread id randomly (%s)." % (
                possible_thread_ids,
                possible_thread_ids[0],
            )

    # If we got here we couldn't discover the main thread id.
    return None, "Unable to discover main thread id."


def fix_main_thread_id(on_warn=lambda msg: None, on_exception=lambda msg: None, on_critical=lambda msg: None):
    # This means that we weren't able to import threading in the main thread (which most
    # likely means that the main thread is paused or in some very long operation).
    # In this case we'll import threading here and hotfix what may be wrong in the threading
    # module (if we're on Windows where we create a thread to do the attach and on Linux
    # we are not certain on which thread we're executing this code).
    #
    # The code below is a workaround for https://bugs.python.org/issue37416
    import sys
    import threading

    # This is no longer needed in Py 3.13 (as the related issue is already fixed).
    if sys.version_info[:2] >= (3, 13):
        return

    try:
        with threading._active_limbo_lock:
            main_thread_instance = get_main_thread_instance(threading)

            if sys.platform == "win32":
                # On windows this code would be called in a secondary thread, so,
                # the current thread is unlikely to be the main thread.
                if hasattr(threading, "_get_ident"):
                    unlikely_thread_id = threading._get_ident()  # py2
                else:
                    unlikely_thread_id = threading.get_ident()  # py3
            else:
                unlikely_thread_id = None

            main_thread_id, critical_warning = get_main_thread_id(unlikely_thread_id)

            if main_thread_id is not None:
                main_thread_id_attr = "_ident"
                if not hasattr(main_thread_instance, main_thread_id_attr):
                    main_thread_id_attr = "_Thread__ident"
                    assert hasattr(main_thread_instance, main_thread_id_attr)

                if main_thread_id != getattr(main_thread_instance, main_thread_id_attr):
                    # Note that we also have to reset the '_tstack_lock' for a regular lock.
                    # This is needed to avoid an error on shutdown because this lock is bound
                    # to the thread state and will be released when the secondary thread
                    # that initialized the lock is finished -- making an assert fail during
                    # process shutdown.
                    main_thread_instance._tstate_lock = threading._allocate_lock()
                    main_thread_instance._tstate_lock.acquire()

                    # Actually patch the thread ident as well as the threading._active dict
                    # (we should have the _active_limbo_lock to do that).
                    threading._active.pop(getattr(main_thread_instance, main_thread_id_attr), None)
                    setattr(main_thread_instance, main_thread_id_attr, main_thread_id)
                    threading._active[getattr(main_thread_instance, main_thread_id_attr)] = main_thread_instance

        # Note: only import from pydevd after the patching is done (we want to do the minimum
        # possible when doing that patching).
        on_warn(
            "The threading module was not imported by user code in the main thread. The debugger will attempt to work around https://bugs.python.org/issue37416."
        )

        if critical_warning:
            on_critical("Issue found when debugger was trying to work around https://bugs.python.org/issue37416:\n%s" % (critical_warning,))
    except:
        on_exception("Error patching main thread id.")


def attach(port, host, protocol="", debug_mode=""):
    try:
        import sys

        fix_main_thread = "threading" not in sys.modules

        if fix_main_thread:

            def on_warn(msg):
                from _pydev_bundle import pydev_log

                pydev_log.warn(msg)

            def on_exception(msg):
                from _pydev_bundle import pydev_log

                pydev_log.exception(msg)

            def on_critical(msg):
                from _pydev_bundle import pydev_log

                pydev_log.critical(msg)

            fix_main_thread_id(on_warn=on_warn, on_exception=on_exception, on_critical=on_critical)

        else:
            from _pydev_bundle import pydev_log  # @Reimport

            pydev_log.debug("The threading module is already imported by user code.")

        if protocol:
            from _pydevd_bundle import pydevd_defaults

            pydevd_defaults.PydevdCustomization.DEFAULT_PROTOCOL = protocol

        if debug_mode:
            from _pydevd_bundle import pydevd_defaults

            pydevd_defaults.PydevdCustomization.DEBUG_MODE = debug_mode

        import pydevd

        # I.e.: disconnect/reset if already connected.

        pydevd.SetupHolder.setup = None

        py_db = pydevd.get_global_debugger()
        if py_db is not None:
            py_db.dispose_and_kill_all_pydevd_threads(wait=False)

        # pydevd.DebugInfoHolder.DEBUG_TRACE_LEVEL = 3
        pydevd.settrace(
            port=port,
            host=host,
            stdout_to_server=True,
            stderr_to_server=True,
            overwrite_prev_trace=True,
            suspend=False,
            trace_only_current_thread=False,
            patch_multiprocessing=False,
        )
    except:
        import traceback

        traceback.print_exc()



import types

from _pydev_bundle import pydev_log
from _pydevd_bundle import pydevd_trace_api
from _pydevd_bundle.pydevd_constants import IS_PYCHARM

try:
    from pydevd_plugins import django_debug
except:
    django_debug = None
    pydev_log.debug('Unable to load django_debug plugin')

try:
    from pydevd_plugins import jinja2_debug
except:
    jinja2_debug = None
    pydev_log.debug('Unable to load jinja2_debug plugin')

try:
    import pydev_jupyter_plugin
except:
    pydev_jupyter_plugin = None
    if IS_PYCHARM:
        pydev_log.debug('Unable to load jupyter_debug plugin')


def load_plugins():
    plugins = []
    if django_debug is not None:
        plugins.append(django_debug)

    if jinja2_debug is not None:
        plugins.append(jinja2_debug)

    if pydev_jupyter_plugin is not None:
        plugins.append(pydev_jupyter_plugin)

    return plugins


def bind_func_to_method(func, obj, method_name):
    bound_method = types.MethodType(func, obj)

    setattr(obj, method_name, bound_method)
    return bound_method


class PluginManager(object):

    def __init__(self, main_debugger):
        self.plugins = load_plugins()
        self.active_plugins = []
        self.main_debugger = main_debugger
        self.rebind_methods()

    def add_breakpoint(self, func_name, *args, **kwargs):
        # add breakpoint for plugin and remember which plugin to use in tracing
        for plugin in self.plugins:
            if hasattr(plugin, func_name):
                func = getattr(plugin, func_name)
                result = func(self, *args, **kwargs)
                if result:
                    self.activate(plugin)

                    return result
        return None

    def activate(self, plugin):
        if plugin not in self.active_plugins:
            self.active_plugins.append(plugin)
            self.rebind_methods()

    def rebind_methods(self):
        if len(self.active_plugins) == 0:
            self.bind_functions(pydevd_trace_api, getattr, pydevd_trace_api)
        elif len(self.active_plugins) == 1:
            self.bind_functions(pydevd_trace_api, getattr, self.active_plugins[0])
        else:
            self.bind_functions(pydevd_trace_api, create_dispatch, self.active_plugins)

    def bind_functions(self, interface, function_factory, arg):
        for name in dir(interface):
            func = function_factory(arg, name)
            if type(func) == types.FunctionType:
                bind_func_to_method(func, self, name)


def create_dispatch(obj, name):
    def dispatch(self, *args, **kwargs):
        result = None
        for p in self.active_plugins:
            r = getattr(p, name)(self, *args, **kwargs)
            if not result:
                result = r
        return result
    return dispatch











# TODO that would make IPython integration better
# - show output other times then when enter was pressed
# - support proper exit to allow IPython to cleanup (e.g. temp files created with %edit)
# - support Ctrl-D (Ctrl-Z on Windows)
# - use IPython (numbered) prompts in PyDev
# - better integration of IPython and PyDev completions
# - some of the semantics on handling the code completion are not correct:
#   eg: Start a line with % and then type c should give %cd as a completion by it doesn't
#       however type %c and request completions and %cd is given as an option
#   eg: Completing a magic when user typed it without the leading % causes the % to be inserted
#       to the left of what should be the first colon.
"""Interface to TerminalInteractiveShell for PyDev Interactive Console frontend
   for IPython 0.11 to 1.0+.
"""

from __future__ import print_function

import os
import sys
import codeop
import traceback
from copy import deepcopy

import IPython
from IPython.core.error import UsageError
from IPython.core.interactiveshell import InteractiveShell, InteractiveShellABC
from IPython.core.usage import default_banner_parts
from IPython.core.formatters import DisplayFormatter
from IPython.core import release

from IPython.terminal.interactiveshell import TerminalInteractiveShell
from IPython.terminal.ipapp import TerminalIPythonApp
from IPython import InteractiveShell

from traitlets import Type
from traitlets import CBool, Unicode


from _pydevd_bundle.pydevd_constants import dict_keys, dict_iter_items
from _pydevd_bundle.pydevd_ipython_console_output import PyDevDebugDisplayHook, PyDevDebugDisplayPub
from _pydev_bundle.pydev_ipython_rich_output import PyDevDisplayHook, PyDevDisplayPub, \
    patch_stdout
from _pydev_bundle.pydev_ipython_completer import init_shell_completer

default_pydev_banner_parts = default_banner_parts
default_pydev_banner = ''.join(default_pydev_banner_parts)

INLINE_OUTPUT_SUPPORTED = os.getenv('INLINE_OUTPUT_SUPPORTED', False)


def show_in_pager(self, strng, *args, **kwargs):
    """ Run a string through pager """
    # On PyDev we just output the string, there are scroll bars in the console
    # to handle "paging". This is the same behaviour as when TERM==dump (see
    # page.py)
    # for compatibility with mime-bundle form:
    if isinstance(strng, dict):
        strng = strng['text/plain']
    print(strng)


def create_editor_hook(rpc_client):
    def call_editor(filename, line=0, wait=True):
        """ Open an editor in PyDev """
        if line is None:
            line = 0

        # Make sure to send an absolution path because unlike most editor hooks
        # we don't launch a process. This is more like what happens in the zmqshell
        filename = os.path.abspath(filename)

        # import sys
        # sys.__stderr__.write('Calling editor at: %s:%s\n' % (pydev_host, pydev_client_port))

        # Tell PyDev to open the editor
        rpc_client.IPythonEditor(filename, str(line))

        if wait:
            try:
                raw_input("Press Enter when done editing:")
            except NameError:
                input("Press Enter when done editing:")
    return call_editor


class PyDevTerminalInteractiveShell(TerminalInteractiveShell):
    banner1 = Unicode(default_pydev_banner, config=True,
        help="""The part of the banner to be printed before the profile"""
    )

    # TODO term_title: (can PyDev's title be changed???, see terminal.py for where to inject code, in particular set_term_title as used by %cd)
    # for now, just disable term_title
    term_title = CBool(False)

    # Note in version 0.11 there is no guard in the IPython code about displaying a
    # warning, so with 0.11 you get:
    #  WARNING: Readline services not available or not loaded.
    #  WARNING: The auto-indent feature requires the readline library
    # Disable readline, readline type code is all handled by PyDev (on Java side)
    readline_use = CBool(False)
    # autoindent has no meaning in PyDev (PyDev always handles that on the Java side),
    # and attempting to enable it will print a warning in the absence of readline.
    autoindent = CBool(False)
    # Force console to not give warning about color scheme choice and default to NoColor.
    # TODO It would be nice to enable colors in PyDev but:
    # - The PyDev Console (Eclipse Console) does not support the full range of colors, so the
    #   effect isn't as nice anyway at the command line
    # - If done, the color scheme should default to LightBG, but actually be dependent on
    #   any settings the user has (such as if a dark theme is in use, then Linux is probably
    #   a better theme).
    colors_force = CBool(True)
    colors = Unicode("nocolor" if IPython.version_info >= (9,) else "NoColor")
    # Since IPython 5 the terminal interface is not compatible with Emacs `inferior-shell` and
    # the `simple_prompt` flag is needed
    simple_prompt = CBool(True)
    pydev_curr_exec_line = 0
    
    if INLINE_OUTPUT_SUPPORTED:
        displayhook_class = Type(PyDevDisplayHook)
        display_pub_class = Type(PyDevDisplayPub)

    def __init__(self, *args, **kwargs):
        super(PyDevTerminalInteractiveShell, self).__init__(*args, **kwargs)
        if INLINE_OUTPUT_SUPPORTED:
            try:
                self.enable_matplotlib('inline')
            except:
                sys.stderr.write("Failed to enable inline matplotlib plots\n")
                sys.stderr.flush()

    def patch_stdout_if_needed(self):
        if INLINE_OUTPUT_SUPPORTED:
            patch_stdout(self)

    # In the PyDev Console, GUI control is done via hookable XML-RPC server
    @staticmethod
    def enable_gui(gui=None, app=None):
        """Switch amongst GUI input hooks by name.
        """
        # Deferred import
        if gui != 'inline':
            from pydev_ipython.inputhook import enable_gui as real_enable_gui
            try:
                return real_enable_gui(gui, app)
            except ValueError as e:
                raise UsageError("%s" % e)

    def init_display_formatter(self):
        if INLINE_OUTPUT_SUPPORTED:
            self.display_formatter = DisplayFormatter(parent=self)
            self.configurables.append(self.display_formatter)
            self.display_formatter.ipython_display_formatter.enabled = True
        else:
            super(PyDevTerminalInteractiveShell, self).init_display_formatter()

    #-------------------------------------------------------------------------
    # Things related to hooks
    #-------------------------------------------------------------------------

    def init_hooks(self):
        super(PyDevTerminalInteractiveShell, self).init_hooks()
        self.set_hook('show_in_pager', show_in_pager)

    #-------------------------------------------------------------------------
    # Things related to exceptions
    #-------------------------------------------------------------------------

    def showtraceback(self, exc_tuple=None, *args, **kwargs):
        # IPython does a lot of clever stuff with Exceptions. However mostly
        # it is related to IPython running in a terminal instead of an IDE.
        # (e.g. it prints out snippets of code around the stack trace)
        # PyDev does a lot of clever stuff too, so leave exception handling
        # with default print_exc that PyDev can parse and do its clever stuff
        # with (e.g. it puts links back to the original source code)
        try:
            if exc_tuple is None:
                etype, value, tb = sys.exc_info()
            else:
                etype, value, tb = exc_tuple
        except ValueError:
            return

        if tb is not None:
            traceback.print_exception(etype, value, tb)
            sys.last_type, sys.last_value, sys.last_traceback = etype, value, tb

    def init_completer(self):
        init_shell_completer(self)

    #-------------------------------------------------------------------------
    # Things related to aliases
    #-------------------------------------------------------------------------

    def init_alias(self):
        # InteractiveShell defines alias's we want, but TerminalInteractiveShell defines
        # ones we don't. So don't use super and instead go right to InteractiveShell
        InteractiveShell.init_alias(self)

    #-------------------------------------------------------------------------
    # Things related to exiting
    #-------------------------------------------------------------------------
    def ask_exit(self):
        """ Ask the shell to exit. Can be overiden and used as a callback. """
        # TODO PyDev's console does not have support from the Python side to exit
        # the console. If user forces the exit (with sys.exit()) then the console
        # simply reports errors. e.g.:
        # >>> import sys
        # >>> sys.exit()
        # Failed to create input stream: Connection refused
        # >>>
        # Console already exited with value: 0 while waiting for an answer.
        # Error stream:
        # Output stream:
        # >>>
        #
        # Alternatively if you use the non-IPython shell this is what happens
        # >>> exit()
        # <type 'exceptions.SystemExit'>:None
        # >>>
        # <type 'exceptions.SystemExit'>:None
        # >>>
        #
        super(PyDevTerminalInteractiveShell, self).ask_exit()
        print('To exit the PyDev Console, terminate the console within IDE.')

    #-------------------------------------------------------------------------
    # Things related to magics
    #-------------------------------------------------------------------------

    def init_magics(self):
        super(PyDevTerminalInteractiveShell, self).init_magics()
        # TODO Any additional magics for PyDev?


class PyDebuggerTerminalInteractiveShell(PyDevTerminalInteractiveShell):
    """
    InteractiveShell for the Jupyter Debug Console.
    Print result outputs to the shell.
    See method 'do_add_exec' from  'pydev/_pydev_bundle/pydev_ipython_code_executor.py'.
    """

    displayhook_class = Type(PyDevDebugDisplayHook)
    display_pub_class = Type(PyDevDebugDisplayPub)
    new_instance = None


InteractiveShellABC.register(PyDevTerminalInteractiveShell)  # @UndefinedVariable
InteractiveShellABC.register(PyDebuggerTerminalInteractiveShell)

class PyDevIpythonApp(TerminalIPythonApp):
    def initialize(self, shell_cls):
        """Do actions after construct, but before starting the app."""
        cl_config = deepcopy(self.config)
        self.init_profile_dir()
        self.init_config_files()
        self.load_config_file()
        self.update_config(cl_config)
        self.init_path()
        self.init_shell(shell_cls)
        self.init_extensions()
        self.init_code()

    def init_shell(self, shell_cls):
        self.shell = shell_cls.instance()
        self.shell.configurables.append(self)


#=======================================================================================================================
# _PyDevIPythonFrontEnd
#=======================================================================================================================
class _PyDevIPythonFrontEnd:

    version = release.__version__

    def __init__(self, is_jupyter_debugger=False):
        # Create and initialize our IPython instance.
        self.is_jupyter_debugger = is_jupyter_debugger
        if is_jupyter_debugger:
            if self._has_shell_instance(PyDebuggerTerminalInteractiveShell, 'new_instance'):
                self.ipython = PyDebuggerTerminalInteractiveShell.new_instance
            else:
                # if we already have some InteractiveConsole instance (Python Console: Attach Debugger)
                if self._has_shell_instance(PyDevTerminalInteractiveShell, '_instance'):
                    PyDevTerminalInteractiveShell.clear_instance()

                InteractiveShell.clear_instance()

                self.ipython = self._init_ipy_app(PyDebuggerTerminalInteractiveShell).shell
                PyDebuggerTerminalInteractiveShell.new_instance = PyDebuggerTerminalInteractiveShell._instance
        else:
            if self._has_shell_instance(PyDevTerminalInteractiveShell, '_instance'):
                self.ipython = PyDevTerminalInteractiveShell._instance
            else:
                self.ipython = self._init_ipy_app(PyDevTerminalInteractiveShell).shell

        self._curr_exec_line = 0
        self._curr_exec_lines = []

    def _init_ipy_app(self, shell_cls):
        application = PyDevIpythonApp()
        application.initialize(shell_cls)
        return application


    def _has_shell_instance(self, shell_cls, instance_str):
        return getattr(shell_cls, instance_str, None) is not None


    def update(self, globals, locals):
        ns = self.ipython.user_ns

        for key in dict_keys(self.ipython.user_ns):
            if key not in locals:
                locals[key] = ns[key]

        self.ipython.user_global_ns.clear()
        self.ipython.user_global_ns.update(globals)

        # If `globals` and `locals` passed to the method are the same objects, we have to ensure that they are also
        # the same in the IPython evaluation context to avoid troubles with some corner-cases such as generator expressions.
        # See: `pydevd_console_integration.console_exec()`.
        self.ipython.user_ns = self.ipython.user_global_ns if globals is locals else locals

        if hasattr(self.ipython, 'history_manager') and getattr(self.ipython.history_manager, 'save_thread', None) is not None:
            self.ipython.history_manager.save_thread.pydev_do_not_trace = True  # don't trace ipython history saving thread

    def complete(self, string):
        try:
            if string:
                return self.ipython.complete(None, line=string, cursor_pos=string.__len__())
            else:
                return self.ipython.complete(string, string, 0)
        except:
            # Silence completer exceptions
            pass

    def is_complete(self, string):
        #Based on IPython 0.10.1

        if string in ('', '\n'):
            # Prefiltering, eg through ipython0, may return an empty
            # string although some operations have been accomplished. We
            # thus want to consider an empty string as a complete
            # statement.
            return True
        else:
            try:
                # Add line returns here, to make sure that the statement is
                # complete (except if '\' was used).
                # This should probably be done in a different place (like
                # maybe 'prefilter_input' method? For now, this works.
                clean_string = string.rstrip('\n')
                if not clean_string.endswith('\\'):
                    clean_string += '\n\n'

                is_complete = codeop.compile_command(
                    clean_string,
                    "<string>",
                    "exec"
                )
            except Exception:
                # XXX: Hack: return True so that the
                # code gets executed and the error captured.
                is_complete = True
            return is_complete

    def getCompletions(self, text, act_tok):
        # Get completions from IPython and from PyDev and merge the results
        # IPython only gives context free list of completions, while PyDev
        # gives detailed information about completions.
        try:
            TYPE_IPYTHON = '11'
            TYPE_IPYTHON_MAGIC = '12'
            _line, ipython_completions = self.complete(text)

            from _pydev_bundle._pydev_completer import Completer
            completer = Completer(self.get_namespace(), None)
            ret = completer.complete(act_tok)
            append = ret.append
            ip = self.ipython
            pydev_completions = set([f[0] for f in ret])
            for ipython_completion in ipython_completions:

                #PyCharm was not expecting completions with '%'...
                #Could be fixed in the backend, but it's probably better
                #fixing it at PyCharm.
                #if ipython_completion.startswith('%'):
                #    ipython_completion = ipython_completion[1:]

                if ipython_completion not in pydev_completions:
                    pydev_completions.add(ipython_completion)
                    inf = ip.object_inspect(ipython_completion)
                    if inf['type_name'] == 'Magic function':
                        pydev_type = TYPE_IPYTHON_MAGIC
                    else:
                        pydev_type = TYPE_IPYTHON
                    pydev_doc = inf['docstring']
                    if pydev_doc is None:
                        pydev_doc = ''
                    append((ipython_completion, pydev_doc, '', pydev_type))
            return ret
        except:
            import traceback;traceback.print_exc()
            return []

    def get_namespace(self):
        return self.ipython.user_ns

    def clear_buffer(self):
        del self._curr_exec_lines[:]

    def add_exec(self, line):
        if self._curr_exec_lines:
            self._curr_exec_lines.append(line)

            buf = '\n'.join(self._curr_exec_lines)

            if self.is_complete(buf):
                self._curr_exec_line += 1
                self.ipython.pydev_curr_exec_line = self._curr_exec_line
                res = self.ipython.run_cell(buf)
                del self._curr_exec_lines[:]
                if res.error_in_exec is not None:
                    return False, True
                else:
                    return False, False #execute complete (no more)

            return True, False #needs more
        else:

            if not self.is_complete(line):
                #Did not execute
                self._curr_exec_lines.append(line)
                return True, False #needs more
            else:
                self._curr_exec_line += 1
                self.ipython.pydev_curr_exec_line = self._curr_exec_line
                if not self.is_jupyter_debugger:
                    res = self.ipython.run_cell(line, store_history=True)
                else:
                    res = self.ipython.run_cell(line, store_history=False)
                if res.error_in_exec is not None:
                    return False, True
                else:
                    return False, False #execute complete (no more)

    def is_automagic(self):
        return self.ipython.automagic

    def get_greeting_msg(self):
        return 'PyDev console: using IPython %s\n' % self.version


class _PyDevFrontEndContainer:
    _instance = None
    _last_rpc_client = None

class _PyDebuggerFrontEndContainer:
    _instance = None

def get_client():
    return _PyDevFrontEndContainer._last_rpc_client


def get_pydev_ipython_frontend(rpc_client, is_jupyter_debugger=False):
    if is_jupyter_debugger:
        if _PyDebuggerFrontEndContainer._instance is None:
            _PyDebuggerFrontEndContainer._instance = _PyDevIPythonFrontEnd(is_jupyter_debugger)

        return _PyDebuggerFrontEndContainer._instance

    if _PyDevFrontEndContainer._instance is None:
        _PyDevFrontEndContainer._instance = _PyDevIPythonFrontEnd(is_jupyter_debugger)

    if _PyDevFrontEndContainer._last_rpc_client != rpc_client:
        _PyDevFrontEndContainer._last_rpc_client = rpc_client

        # Back channel to PyDev to open editors (in the future other
        # info may go back this way. This is the same channel that is
        # used to get stdin, see StdIn in pydev_console_utils)
        _PyDevFrontEndContainer._instance.ipython.hooks['editor'] = create_editor_hook(rpc_client)

        # Note: setting the callback directly because setting it with set_hook would actually create a chain instead
        # of ovewriting at each new call).
        # _PyDevFrontEndContainer._instance.ipython.set_hook('editor', create_editor_hook(pydev_host, pydev_client_port))

    return _PyDevFrontEndContainer._instance


def get_ipython_hidden_vars(ipython_shell):
    try:
        if hasattr(ipython_shell, 'user_ns_hidden'):
            user_ns_hidden = ipython_shell.user_ns_hidden
            if isinstance(user_ns_hidden, dict):
                # Since IPython 2 dict `user_ns_hidden` contains hidden variables and values
                user_hidden_dict = user_ns_hidden.copy()
            else:
                # In IPython 1.x `user_ns_hidden` used to be a set with names of hidden variables
                user_hidden_dict = dict([(key, val) for key, val in dict_iter_items(ipython_shell.user_ns)
                                         if key in user_ns_hidden])

            # while `_`, `__` and `___` were not initialized, they are not presented in `user_ns_hidden`
            user_hidden_dict.setdefault('_', '')
            user_hidden_dict.setdefault('__', '')
            user_hidden_dict.setdefault('___', '')

            return user_hidden_dict
    except:
        # Getting IPython variables shouldn't break loading frame variables
        traceback.print_exc()



#  Copyright 2000-2024 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
import dis
from opcode import HAVE_ARGUMENT, opname, EXTENDED_ARG, hasconst, hasname, hasjrel, \
    haslocal, hascompare, cmp_op, hasfree

from _pydevd_bundle.smart_step_into.pydevd_smart_step_into_util import Instruction


def _is_call_opname(opname):
    return opname.startswith('CALL_')


_LOAD_OPNAMES = {
    'LOAD_GLOBAL',
    'LOAD_ATTR',
    'LOAD_FAST',
    'LOAD_NAME',
    'LOAD_DEREF',
    'LOAD_CLOSURE',
}


def _is_load_opname(opname):
    return opname in _LOAD_OPNAMES


def get_stepping_variants(code):
    stk = []
    for instruction in get_instructions(code):
        if _is_call_opname(instruction.opname):
            while stk and not _is_load_opname(stk[-1].opname):
                stk.pop()
            if not stk:
                continue
            tos = stk.pop()
            yield Instruction(
                tos.opname,
                tos.opcode,
                tos.arg,
                tos.argval,
                instruction.offset,
                tos.lineno
            )
        else:
            stk.append(instruction)


def get_instructions(co):
    code = co.co_code
    linestarts = dict(dis.findlinestarts(co))
    curr_line = -1
    n = len(code)
    i = 0
    extended_arg = 0
    free = None
    while i < n:
        c = code[i]
        op = ord(c)
        if i in linestarts:
            curr_line = linestarts[i]
        offset = i
        i = i + 1
        oparg = argval = ''
        if op >= HAVE_ARGUMENT:
            oparg = ord(code[i]) + ord(code[i + 1]) * 256 + extended_arg
            extended_arg = 0
            i = i+2
            if op == EXTENDED_ARG:
                extended_arg = oparg*65536L
            if op in hasconst:
                argval = repr(co.co_consts[oparg])
            elif op in hasname:
                argval = co.co_names[oparg]
            elif op in hasjrel:
                argval = repr(i + oparg)
            elif op in haslocal:
                argval = co.co_varnames[oparg]
            elif op in hascompare:
                argval = cmp_op[oparg]
            elif op in hasfree:
                if free is None:
                    free = co.co_cellvars + co.co_freevars
                argval = free[oparg]
        yield Instruction(
            opname[op],
            op,
            oparg,
            argval,
            offset,
            curr_line
        )



import os
import sys

from _pydev_bundle.pydev_monkey import log_error_once
from _pydevd_bundle.pydevd_constants import IS_PYCHARM, IS_PY311, IS_PY312_OR_GREATER

IS_PY36_OR_GREATER = sys.version_info >= (3, 6)

frame_eval_func = None
stop_frame_eval = None
dummy_trace_dispatch = None
show_frame_eval_warning = False
clear_thread_local_info = None

# "NO" means we should not use frame evaluation, 'YES' we should use it (and fail if not there) and unspecified uses if possible.
use_frame_eval = os.environ.get('PYDEVD_USE_FRAME_EVAL', None)
use_cython = os.getenv('PYDEVD_USE_CYTHON', None)


if not IS_PY36_OR_GREATER or sys.version_info[:3] == (3, 6, 1):  # PY-37312
    pass

elif IS_PY311:  # PY-51730
    pass

elif IS_PY312_OR_GREATER:  # PEP 669 tracing should be used instead.
    pass

elif use_cython == 'NO':
    log_error_once("warning: PYDEVD_USE_CYTHON environment variable is set to 'NO'. "
                   "Frame evaluator will be also disabled because it requires Cython extensions to be enabled in order to operate correctly.")

else:
    if use_frame_eval == 'NO':
        pass

    elif use_frame_eval == 'YES':
        try:
            import _pydevd_bundle.pydevd_cython_wrapper
        except ImportError:
            # Frame evaluator doesn't work without the Cython speedups.
            pass
        else:
            # Fail if unable to use
            from _pydevd_frame_eval.pydevd_frame_eval_cython_wrapper import frame_eval_func, stop_frame_eval, dummy_trace_dispatch, clear_thread_local_info

    elif use_frame_eval is None:
        # Try to use if possible
        try:
            try:
                import _pydevd_bundle.pydevd_cython_wrapper
            except ImportError:
                # Frame evaluator doesn't work without the Cython speedups.
                pass
            else:
                from _pydevd_frame_eval.pydevd_frame_eval_cython_wrapper import frame_eval_func, stop_frame_eval, dummy_trace_dispatch, clear_thread_local_info
        except ImportError:
            dirname = os.path.dirname(os.path.dirname(__file__))
            if not IS_PYCHARM:
                log_error_once("warning: Debugger speedups using cython not found. Run '\"%s\" \"%s\" build_ext --inplace' to build." % (
                    sys.executable, os.path.join(dirname, 'setup_cython.py')))
            else:
                show_frame_eval_warning = True

    else:
        raise RuntimeError('Unexpected value for PYDEVD_USE_FRAME_EVAL: %s (accepted: YES, NO)' % (use_frame_eval,))



"""
A simpler setup version just to compile the speedup module.

It should be used as:

python setup_cython build_ext --inplace

Note: the .c file and other generated files are regenerated from
the .pyx file by running "python build_tools/build.py"
"""
from __future__ import with_statement

import os
import shutil
import sys
from setuptools import setup

_pydevd_dir = os.path.dirname(os.path.abspath(__file__))
os.chdir(_pydevd_dir)

IS_PY36_OR_GREATER = sys.version_info > (3, 6)
IS_PY311_OR_GREATER = sys.version_info > (3, 11)
IS_PY312_OR_GREATER = sys.version_info > (3, 12)


def process_args():
    extension_folder = None
    target_pydevd_name = None
    target_frame_eval = None
    force_cython = False
    target_arch = None

    for i, arg in enumerate(sys.argv[:]):
        if arg == '--build-lib':
            extension_folder = sys.argv[i + 1]
            # It shouldn't be removed from sys.argv (among with --build-temp)
            # because they're passed further to setup().
        if arg.startswith('--target-pyd-name='):
            sys.argv.remove(arg)
            target_pydevd_name = arg[len('--target-pyd-name='):]
        if arg.startswith('--target-pyd-frame-eval='):
            sys.argv.remove(arg)
            target_frame_eval = arg[len('--target-pyd-frame-eval='):]
        if arg == '--force-cython':
            sys.argv.remove(arg)
            force_cython = True
        if arg.startswith('--target='):
            sys.argv.remove(arg)
            target_arch = arg[len('--target='):]

    return extension_folder, target_pydevd_name, target_frame_eval, \
        force_cython, target_arch


def build_extension(dir_name, extension_name, target_pydevd_name, force_cython,
                    target_arch, extended=False, should_add_python_3_12_suffix=True):
    """
    :param dir_name: directory where the Cython file is located
    :param extension_name: name of the .pyx or .c file to build extension from, e.g.
      'pydevd_cython' or 'pydevd_frame_evaluator_39_310'
    :param target_pydevd_name: name of the resulting extension
    :param force_cython: if False, build from the C file, use ``.pyx`` otherwise
    :param target_arch: target architecture, e.g. amd64
    :param extended: add ``_ext`` to the name of the extension package name
    :param should_add_python_3_12_suffix: whether to add a Python 3.12 specific suffixes
      to the build files
    """
    pyx_file = os.path.join(os.path.dirname(__file__), dir_name,
                            "%s.pyx" % (extension_name,))
    has_pxd = False

    if target_pydevd_name != extension_name:
        # It MUST be there in this case! (Otherwise we'll have unresolved externals
        # because the .c file had another name initially).

        # We must force Cython in this case (but only in this case -- for the regular
        # setup in the user machine, we should always compile the .c file).
        force_cython = True

        new_pyx_file = os.path.join(os.path.dirname(__file__), dir_name,
                                    "%s.pyx" % (target_pydevd_name,))
        new_c_file = os.path.join(os.path.dirname(__file__), dir_name,
                                  "%s.c" % (target_pydevd_name,))
        shutil.copy(pyx_file, new_pyx_file)
        pyx_file = new_pyx_file

        pxd_file = os.path.join(os.path.dirname(__file__), dir_name,
                                "%s.pxd" % (extension_name,))
        if os.path.exists(pxd_file) and os.path.isfile(pxd_file):
            new_pxd_file = os.path.join(os.path.dirname(__file__), dir_name,
                                        "%s.pxd" % (target_pydevd_name,))
            shutil.copy(pxd_file, new_pxd_file)
            has_pxd = True

        assert os.path.exists(pyx_file)
        if has_pxd:
            assert os.path.exists(new_pxd_file)

    try:
        from distutils.extension import Extension

        if target_arch:
            extra_compile_args = ["--target=%s" % target_arch]
            extra_link_args = ["--target=%s" % target_arch]
        else:
            extra_compile_args = []
            extra_link_args = []

        if force_cython:
            # noinspection PyPackageRequirements
            from Cython.Build import cythonize
            ext_modules = cythonize(Extension(
                "%s.%s" % (dir_name, target_pydevd_name,),
                ["%s/%s.pyx" % (dir_name, target_pydevd_name,)],
                extra_compile_args=extra_compile_args,
                extra_link_args=extra_link_args,
                ),
                # See: https://cython.readthedocs.io/en/latest/src/userguide/migrating_to_cy30.html#binding-functions
                compiler_directives={'binding': False},
                force=True)
        else:
            # Always compile the .c (and not the .pyx) file (which we should keep
            # up-to-date by running build_tools/build.py).
            ext_modules = [Extension(
                "%s%s.%s" % (dir_name, "_ext" if extended else "", target_pydevd_name),
                [os.path.join(dir_name, extension_name + ".c")],
                # uncomment to generate pdbs for visual studio.
                # extra_compile_args=["-Zi", "/Od"],
                # extra_link_args=["-debug"],
                extra_compile_args=extra_compile_args,
                extra_link_args=extra_link_args,
                )]

        setup(
            name="Cythonize",
            ext_modules=ext_modules
        )
    finally:
        if target_pydevd_name != extension_name:
            try:
                # noinspection PyUnboundLocalVariable
                os.remove(new_pyx_file)
            except:  # noqa: 722
                import traceback
                traceback.print_exc()
            try:
                # noinspection PyUnboundLocalVariable
                os.remove(new_c_file)
            except: # noqa: 722
                import traceback
                traceback.print_exc()
            if has_pxd:
                try:
                    # noinspection PyUnboundLocalVariable
                    os.remove(new_pxd_file)
                except:  # noqa: 722
                    import traceback
                    traceback.print_exc()


def get_frame_eval_extension_name():
    version = sys.version_info[:2]
    if (3, 6) <= version <= (3, 8):
        return "pydevd_frame_evaluator_36_38"
    elif (3, 9) <= version <= (3, 10):
        return "pydevd_frame_evaluator_39_310"
    else:
        raise RuntimeError("Frame evaluation is not supported for the Python version"
                           "%s.%s" % (version[0], version[1]))


def create_init_py_files(extension_folder, subdir_names_to_ignore=None):
    """
    Create `__init__.py` files in every subdirectory of :param:`extension_folder` to
    make sure Python 2 will be able to import from them.
    """
    os.chdir(extension_folder)
    subdir_names_to_ignore = subdir_names_to_ignore or []
    for folder in (f for f in os.listdir(extension_folder)
                   if f not in subdir_names_to_ignore
                      and os.path.isdir(os.path.join(extension_folder, f))):
        file = os.path.join(folder, "__init__.py")
        if not os.path.exists(file):
            open(file, 'a').close()


def main():
    extension_folder, target_pydevd_name, target_frame_eval, force_cython, target_arch \
        = process_args()

    extended = bool(extension_folder)

    extension_name = "pydevd_cython"
    target_pydevd_name = target_pydevd_name or extension_name

    build_extension(
        "_pydevd_bundle", extension_name, target_pydevd_name, force_cython, target_arch,
        extended)

    if IS_PY36_OR_GREATER and not IS_PY311_OR_GREATER:
        extension_name = get_frame_eval_extension_name()
        frame_eval_dir_name = "_pydevd_frame_eval"
        target_frame_eval = target_frame_eval or extension_name

        build_extension(frame_eval_dir_name, extension_name, target_frame_eval,
                        force_cython, target_arch, extended)

    if IS_PY312_OR_GREATER:
        extension_name = "pydevd_pep_669_tracing_cython"
        build_extension("_pydevd_bundle", extension_name, extension_name,
                        force_cython, target_arch, extended, False)

    if extension_folder:
        create_init_py_files(extension_folder, subdir_names_to_ignore=["build"])


if __name__ == "__main__":
    main()



import os
import sys

from _pydevd_bundle.pydevd_constants import IS_PY312_OR_GREATER

use_cython = os.getenv('PYDEVD_USE_CYTHON', None)

if use_cython == 'NO':
    from _pydevd_bundle import pydevd_pep_669_tracing as mod
elif use_cython == 'YES' and IS_PY312_OR_GREATER:
    from _pydevd_bundle import pydevd_pep_669_tracing_cython as mod
else:
    try:
        from _pydevd_bundle import pydevd_pep_669_tracing_cython as mod
    except ImportError:
        from _pydevd_bundle import pydevd_pep_669_tracing as mod

sys.modules['_pydevd_bundle.pydevd_pep_669_tracing'] = mod

enable_pep669_monitoring = mod.enable_pep669_monitoring
disable_pep669_monitoring = mod.disable_pep669_monitoring
global_cache_skips = mod.global_cache_skips
global_cache_frame_skips = mod.global_cache_frame_skips

def _dummy_restart_events():
    pass

try:
    restart_events = sys.monitoring.restart_events
except AttributeError:
    restart_events = _dummy_restart_events



WRITE_OUT = 1
WRITE_ERR = 2
MAX_STRING_LENGTH = 1000


class ConsoleOutputHook:
    def __init__(self, dbg, out, is_stderr=False):
        self.dbg = dbg
        self.original_out = out
        self.is_error = is_stderr

    def write(self, str):
        if len(str) > MAX_STRING_LENGTH:
            self._write_long_string(str)
        else:
            self._write_short_string(str)

    def flush(self):
        pass

    def __getattr__(self, item):
        # it's called if the attribute wasn't found
        if hasattr(self.original_out, item):
            return getattr(self.original_out, item)
        raise AttributeError("%s has no attribute %s" % (self.original_out, item))

    def _write_long_string(self, str):
        str_len, chunk_size = len(str), len(str) // MAX_STRING_LENGTH
        for i in range(0, str_len, chunk_size):
            self._write_short_string(str[i:i + chunk_size])

    def _write_short_string(self, str):
        if self.is_error:
            cmd = self.dbg.cmd_factory.make_io_message(str, WRITE_ERR)
        else:
            cmd = self.dbg.cmd_factory.make_io_message(str, WRITE_OUT)

        self.dbg.writer.add_command(cmd)



#  Copyright 2000-2023 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
import os

from generation_utils import create_temp_dir, get_builtin_modules_name, \
    create_temp_file, clear_temp_dir, write_result_to_file, debug_file

START_GEN_STRING = 'Start generation for %s'
GEN_FINISHED_STRING = 'Generation finished for %s'


def main():
    create_temp_dir()
    try:
        builtin_modules = get_builtin_modules_name()
        old_env = os.environ.copy()
        for module_name in builtin_modules:
            try:
                module_name += '.py'
                create_temp_file(module_name)
                print(START_GEN_STRING % module_name)
                debug_file(module_name)
            except Exception as e:
                print(e)
            finally:
                print(GEN_FINISHED_STRING % module_name)
                clear_temp_dir()
                os.environ = old_env
    except Exception as e:
        print(e)

    write_result_to_file()


if __name__ == '__main__':
    main()



import sys
import traceback
from types import ModuleType


class ImportHookManager(ModuleType):
    def __init__(self, name, system_import):
        ModuleType.__init__(self, name)
        self._system_import = system_import
        self._modules_to_patch = {}
        self.inside_activation = False

    def add_module_name(self, module_name, activate_function):
        self._modules_to_patch[module_name] = activate_function

    def do_import(self, name, *args, **kwargs):
        activate_func = None
        if name in self._modules_to_patch:
            activate_func = self._modules_to_patch[name]

        module = self._system_import(name, *args, **kwargs)
        try:
            if activate_func and not self.inside_activation:
                self.inside_activation = True
                succeeded = activate_func()
                if succeeded and name in self._modules_to_patch:
                    # Remove if only it was executed correctly
                    self._modules_to_patch.pop(name)
                self.inside_activation = False
        except:
            sys.stderr.write("Matplotlib support failed\n")
            traceback.print_exc()
        return module


if sys.version_info[0] >= 3:
    import builtins  # py3
else:
    import __builtin__ as builtins

import_hook_manager = ImportHookManager(__name__ + '.import_hook', builtins.__import__)
builtins.__import__ = import_hook_manager.do_import
sys.modules[import_hook_manager.__name__] = import_hook_manager
del builtins



from __future__ import unicode_literals

import os

from build import remove_binaries
from build_tools.build_common import regenerate_binaries, ensure_interpreters

miniconda64_envs = os.getenv('MINICONDA64_ENVS')
python_installations = [
    r'%s/py27_64/bin/python' % miniconda64_envs,
    r'%s/py36_64/bin/python' % miniconda64_envs,
    r'%s/py37_64/bin/python' % miniconda64_envs,
    r'%s/py38_64/bin/python' % miniconda64_envs,
    r'%s/py39_64/bin/python' % miniconda64_envs,
    r'%s/py310_64/bin/python' % miniconda64_envs,
    r'%s/py311_64/bin/python' % miniconda64_envs,
    r'%s/py312_64/bin/python' % miniconda64_envs,
    ]


def main():
    from generate_code import generate_dont_trace_files
    from generate_code import generate_cython_module

    # First, make sure that our code is up-to-date.
    generate_dont_trace_files()
    generate_cython_module()

    ensure_interpreters(python_installations)

    remove_binaries(['.so'])

    regenerate_binaries(python_installations)


if __name__ == '__main__':
    main()
